# -*- coding: UTF-8 -*-

# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import unittest

from random import randint

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from ripe.atlas.tools.commands.measure.base import Command
from ripe.atlas.tools.commands.measure import (
    PingMeasureCommand,
    TracerouteMeasureCommand,
    DnsMeasureCommand,
    SslcertMeasureCommand,
    HttpMeasureCommand,
    NtpMeasureCommand,
)
from ripe.atlas.tools.exceptions import RipeAtlasToolsException
from ripe.atlas.tools.settings import Configuration, AliasesDB

from ..base import capture_sys_output


class TestMeasureCommand(unittest.TestCase):

    CONF = "ripe.atlas.tools.commands.measure.base.conf"
    KINDS = {
        "ping": PingMeasureCommand,
        "traceroute": TracerouteMeasureCommand,
        "dns": DnsMeasureCommand,
        "sslcert": SslcertMeasureCommand,
        "http": HttpMeasureCommand,
        "ntp": NtpMeasureCommand,
    }

    def setUp(self):
        self.maxDiff = None

    def test_no_arguments(self):
        with capture_sys_output():
            with self.assertRaises(RipeAtlasToolsException) as e:
                Command().init_args([])
            self.assertTrue(
                str(e.exception).startswith("Usage: ripe-atlas measure <"))

    def test_bad_type_argument(self):
        with capture_sys_output():
            with self.assertRaises(RipeAtlasToolsException) as e:
                Command().init_args(["not-a-type"])
            self.assertTrue(
                str(e.exception).startswith("Usage: ripe-atlas measure <"))

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_dry_run(self):

        with capture_sys_output() as (stdout, stderr):
            cmd = PingMeasureCommand()
            cmd.init_args(["ping", "--target", "ripe.net", "--dry-run"])
            cmd.run()
            expected = (
                "\n"
                "Definitions:\n"
                "================================================================================\n"
                "target                    ripe.net\n"
                "packet_interval           1000\n"
                "description               Ping measurement to ripe.net\n"
                "af                        4\n"
                "packets                   3\n"
                "size                      48\n"
                "\n"
                "Sources:\n"
                "================================================================================\n"
                "requested                 50\n"
                "type                      area\n"
                "value                     WW\n"
                "tags\n"
                "  include                 system-ipv4-works\n"
                "  exclude                 \n"
                "\n"
            )
            self.assertEqual(
                set(stdout.getvalue().split("\n")),
                set(expected.split("\n"))
            )

        with capture_sys_output() as (stdout, stderr):
            cmd = PingMeasureCommand()
            cmd.init_args([
                "ping",
                "--target", "ripe.net",
                "--af", "6",
                "--include-tag", "alpha",
                "--include-tag", "bravo",
                "--include-tag", "charlie",
                "--exclude-tag", "delta",
                "--exclude-tag", "echo",
                "--exclude-tag", "foxtrot",
                "--dry-run"
            ])
            cmd.run()
            expected = (
                "\n"
                "Definitions:\n"
                "================================================================================\n"
                "target                    ripe.net\n"
                "packet_interval           1000\n"
                "description               Ping measurement to ripe.net\n"
                "af                        6\n"
                "packets                   3\n"
                "size                      48\n"
                "\n"
                "Sources:\n"
                "================================================================================\n"
                "requested                 50\n"
                "type                      area\n"
                "value                     WW\n"
                "tags\n"
                "  include                 alpha, bravo, charlie\n"
                "  exclude                 delta, echo, foxtrot\n"
                "\n"
            )
            self.assertEqual(
                set(stdout.getvalue().split("\n")),
                set(expected.split("\n"))
            )

    def test_clean_target(self):

        cmd = PingMeasureCommand()
        with capture_sys_output():
            cmd.init_args(["ping", "--target", "ripe.net"])
            self.assertEqual(cmd.clean_target(), "ripe.net")

        cmd = DnsMeasureCommand()
        with capture_sys_output():
            cmd.init_args(["dns", "--query-argument", "ripe.net"])
            self.assertEqual(cmd.clean_target(), None)

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_ping(self):

        spec = Configuration.DEFAULT["specification"]["types"]["ping"]

        cmd = PingMeasureCommand()
        cmd.init_args([
            "ping", "--target", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Ping measurement to ripe.net",
                "target": "ripe.net",
                "packets": spec["packets"],
                "packet_interval": spec["packet-interval"],
                "size": spec["size"]
            }
        )

        cmd = PingMeasureCommand()
        cmd.init_args([
            "ping",
            "--target", "ripe.net",
            "--af", "6",
            "--description", "This is my description",
            "--packets", "7",
            "--packet-interval", "200",
            "--size", "24",
            "--group-id", "1000001",
            "--include-probe-id",
            "--spread", "60",
            "--resolve-on-probe",
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "ripe.net",
                "packets": 7,
                "packet_interval": 200,
                "size": 24,
                "group_id": 1000001,
                "include_probe_id": True,
                "spread": 60,
                "resolve_on_probe": True,
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_traceroute(self):

        spec = Configuration.DEFAULT["specification"]["types"]["traceroute"]

        cmd = TracerouteMeasureCommand()
        cmd.init_args([
            "traceroute", "--target", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Traceroute measurement to ripe.net",
                "target": "ripe.net",
                "packets": spec["packets"],
                "size": spec["size"],
                "destination_option_size": spec["destination-option-size"],
                "hop_by_hop_option_size": spec["hop-by-hop-option-size"],
                "dont_fragment": spec["dont-fragment"],
                "first_hop": spec["first-hop"],
                "max_hops": spec["max-hops"],
                "paris": spec["paris"],
                "port": spec["port"],
                "protocol": spec["protocol"],
                "timeout": spec["timeout"]
            }
        )

        cmd = TracerouteMeasureCommand()
        cmd.init_args([
            "traceroute",
            "--af", "6",
            "--description", "This is my description",
            "--target", "ripe.net",
            "--packets", "7",
            "--size", "24",
            "--destination-option-size", "12",
            "--hop-by-hop-option-size", "13",
            "--dont-fragment",
            "--first-hop", "2",
            "--max-hops", "5",
            "--paris", "8",
            "--port", "123",
            "--protocol", "TCP",
            "--timeout", "1500",
            "--duplicate-timeout", "500",
            "--response-timeout", "500",
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "ripe.net",
                "packets": 7,
                "size": 24,
                "destination_option_size": 12,
                "hop_by_hop_option_size": 13,
                "dont_fragment": True,
                "first_hop": 2,
                "max_hops": 5,
                "paris": 8,
                "port": 123,
                "protocol": "TCP",
                "timeout": 1500,
                "duplicate_timeout": 500,
                "response_timeout": 500,
            }
        )

        """Tests for the protocol provided in lower case"""
        cmd = TracerouteMeasureCommand()
        cmd.init_args([
            "traceroute", "--target", "ripe.net", "--protocol", "icmp"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Traceroute measurement to ripe.net",
                "target": "ripe.net",
                "packets": spec["packets"],
                "size": spec["size"],
                "destination_option_size": spec["destination-option-size"],
                "hop_by_hop_option_size": spec["hop-by-hop-option-size"],
                "dont_fragment": spec["dont-fragment"],
                "first_hop": spec["first-hop"],
                "max_hops": spec["max-hops"],
                "paris": spec["paris"],
                "port": spec["port"],
                "protocol": "ICMP",
                "timeout": spec["timeout"]
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_dns(self):

        spec = Configuration.DEFAULT["specification"]["types"]["dns"]

        cmd = DnsMeasureCommand()
        cmd.init_args([
            "dns", "--query-argument", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "DNS measurement for ripe.net",
                "query_class": spec["query-class"],
                "query_type": spec["query-type"],
                "query_argument": "ripe.net",
                "set_cd_bit": spec["set-cd-bit"],
                "set_do_bit": spec["set-do-bit"],
                "set_rd_bit": spec["set-rd-bit"],
                "set_nsid_bit": spec["set-nsid-bit"],
                "protocol": spec["protocol"],
                "retry": spec["retry"],
                "udp_payload_size": spec["udp-payload-size"],
                "use_probe_resolver": True,
            }
        )

        cmd = DnsMeasureCommand()
        cmd.init_args([
            "dns",
            "--af", "6",
            "--description", "This is my description",
            "--target", "non-existent-dns.ripe.net",
            "--query-class", "CHAOS",
            "--query-type", "SOA",
            "--query-argument", "some arbitrary string",
            "--set-cd-bit",
            "--set-do-bit",
            "--set-rd-bit",
            "--set-nsid-bit",
            "--protocol", "TCP",
            "--retry", "2",
            "--udp-payload-size", "512",
            "--timeout", "500",
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "non-existent-dns.ripe.net",
                "query_class": "CHAOS",
                "query_type": "SOA",
                "query_argument": "some arbitrary string",
                "set_cd_bit": True,
                "set_do_bit": True,
                "set_rd_bit": True,
                "set_nsid_bit": True,
                "protocol": "TCP",
                "retry": 2,
                "udp_payload_size": 512,
                "use_probe_resolver": False,
                "timeout": 500,
            }
        )

        """Testing for protocol argument in lower case"""
        cmd = DnsMeasureCommand()
        cmd.init_args([
            "dns", "--query-argument", "ripe.net", "--protocol", "tcp"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "DNS measurement for ripe.net",
                "query_class": spec["query-class"],
                "query_type": spec["query-type"],
                "query_argument": "ripe.net",
                "set_cd_bit": spec["set-cd-bit"],
                "set_do_bit": spec["set-do-bit"],
                "set_rd_bit": spec["set-rd-bit"],
                "set_nsid_bit": spec["set-nsid-bit"],
                "protocol": "TCP",
                "retry": spec["retry"],
                "udp_payload_size": spec["udp-payload-size"],
                "use_probe_resolver": True,
            }
        )

        """Testing for query class in lower case"""
        cmd = DnsMeasureCommand()
        cmd.init_args([
            "dns", "--query-argument", "ripe.net", "--query-class", "in"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "DNS measurement for ripe.net",
                "query_class": "IN",
                "query_type": spec["query-type"],
                "query_argument": "ripe.net",
                "set_cd_bit": spec["set-cd-bit"],
                "set_do_bit": spec["set-do-bit"],
                "set_rd_bit": spec["set-rd-bit"],
                "set_nsid_bit": spec["set-nsid-bit"],
                "protocol": spec["protocol"],
                "retry": spec["retry"],
                "udp_payload_size": spec["udp-payload-size"],
                "use_probe_resolver": True,
            }
        )

        """Testing for query type in lower case"""
        cmd = DnsMeasureCommand()
        cmd.init_args([
            "dns", "--query-argument", "ripe.net", "--query-type", "txt",
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "DNS measurement for ripe.net",
                "query_class": spec["query-class"],
                "query_type": "TXT",
                "query_argument": "ripe.net",
                "set_cd_bit": spec["set-cd-bit"],
                "set_do_bit": spec["set-do-bit"],
                "set_rd_bit": spec["set-rd-bit"],
                "set_nsid_bit": spec["set-nsid-bit"],
                "protocol": spec["protocol"],
                "retry": spec["retry"],
                "udp_payload_size": spec["udp-payload-size"],
                "use_probe_resolver": True,
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_sslcert(self):

        spec = Configuration.DEFAULT["specification"]["types"]["sslcert"]

        cmd = SslcertMeasureCommand()
        cmd.init_args([
            "sslcert", "--target", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Sslcert measurement to ripe.net",
                "target": "ripe.net",
                "port": spec["port"]
            }
        )

        cmd = SslcertMeasureCommand()
        cmd.init_args([
            "sslcert",
            "--target", "ripe.net",
            "--af", "6",
            "--description", "This is my description",
            "--port", "7",
            "--hostname", "ripe.net",
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "ripe.net",
                "port": 7,
                "hostname": "ripe.net",
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_http(self):

        spec = Configuration.DEFAULT["specification"]["types"]["http"]

        cmd = HttpMeasureCommand()
        cmd.init_args([
            "http", "--target", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Http measurement to ripe.net",
                "target": "ripe.net",
                "header_bytes": spec["header-bytes"],
                "version": spec["version"],
                "method": spec["method"],
                "port": spec["port"],
                "path": spec["path"],
                "query_string": spec["query-string"],
                "user_agent": spec["user-agent"],
                "max_bytes_read": spec["body-bytes"],
            }
        )

        cmd = HttpMeasureCommand()
        cmd.init_args([
            "http",
            "--target", "ripe.net",
            "--af", "6",
            "--description", "This is my description",
            "--header-bytes", "100",
            "--version", "99",
            "--method", "a-method",
            "--port", "7",
            "--path", "/path/to/something",
            "--query-string", "x=7",
            "--user-agent", "This is my custom user agent",
            "--body-bytes", "200",
            "--timing-verbosity", "2"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "ripe.net",
                "header_bytes": 100,
                "version": "99",
                "method": "a-method",
                "port": 7,
                "path": "/path/to/something",
                "query_string": "x=7",
                "user_agent": "This is my custom user agent",
                "max_bytes_read": 200,
                "extended_timing": True,
                "more_extended_timing": True
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_measurement_kwargs_ntp(self):

        spec = Configuration.DEFAULT["specification"]["types"]["ntp"]

        cmd = NtpMeasureCommand()
        cmd.init_args([
            "ntp", "--target", "ripe.net"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": Configuration.DEFAULT["specification"]["af"],
                "description": "Ntp measurement to ripe.net",
                "target": "ripe.net",
                "packets": spec["packets"],
                "timeout": spec["timeout"]
            }
        )

        cmd = NtpMeasureCommand()
        cmd.init_args([
            "ntp",
            "--target", "ripe.net",
            "--af", "6",
            "--description", "This is my description",
            "--packets", "6",
            "--timeout", "9000"
        ])
        self.assertEqual(
            cmd._get_measurement_kwargs(),
            {
                "af": 6,
                "description": "This is my description",
                "target": "ripe.net",
                "packets": 6,
                "timeout": 9000,
            }
        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_get_source_kwargs(self):

        spec = Configuration.DEFAULT["specification"]

        for kind, klass in self.KINDS.items():

            cmd = klass()
            args = [kind, "--target", "example.com"]
            if kind == "dns":
                args = ["dns", "--query-argument", "example.com"]

            cmd.init_args(list(args))

            tags = spec["tags"]["ipv{}".format(cmd._get_af())]
            includes = tags[kind]["include"] + tags["all"]["include"]
            excludes = tags[kind]["exclude"] + tags["all"]["exclude"]

            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": spec["source"]["requested"],
                "type": spec["source"]["type"],
                "value": spec["source"]["value"],
                "tags": {"include": includes, "exclude": excludes}
            })

            cmd = klass()
            cmd.init_args(args + [
                "--probes", "10",
                "--from-country", "ca"
            ])
            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": 10,
                "type": "country",
                "value": "CA",
                "tags": {"include": includes, "exclude": excludes}
            })

            for area in ("WW", "West", "North-Central", "South-Central",
                         "North-East", "South-East"):
                cmd = klass()
                cmd.init_args(args + [
                    "--probes", "10",
                    "--from-area", area
                ])
                self.assertEqual(cmd._get_source_kwargs(), {
                    "requested": 10,
                    "type": "area",
                    "value": area,
                    "tags": {"include": includes, "exclude": excludes}
                })

            cmd = klass()
            cmd.init_args(args + [
                "--probes", "10",
                "--from-prefix", "1.2.3.0/22"
            ])
            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": 10,
                "type": "prefix",
                "value": "1.2.3.0/22",
                "tags": {"include": includes, "exclude": excludes}
            })

            cmd = klass()
            cmd.init_args(args + [
                "--probes", "10",
                "--from-asn", "3333"
            ])
            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": 10,
                "type": "asn",
                "value": 3333,
                "tags": {"include": includes, "exclude": excludes}
            })

            # Try 20 permutations of probe lists
            for __ in range(0, 20):
                requested = randint(1, 500)
                selection = [str(randint(1, 5000)) for _ in range(0, 100)]
                cmd = klass()
                cmd.init_args(args + [
                    "--probes", str(requested),
                    "--from-probes", ",".join(selection)
                ])
                self.assertEqual(cmd._get_source_kwargs(), {
                    "requested": requested,
                    "type": "probes",
                    "value": ",".join(selection),
                    "tags": {"include": includes, "exclude": excludes}
                })

            cmd = klass()
            cmd.init_args(args + [
                "--probes", "10",
                "--from-measurement", "1001"
            ])
            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": 10,
                "type": "msm",
                "value": 1001,
                "tags": {"include": includes, "exclude": excludes}
            })

            cmd = klass()
            cmd.init_args(args + [
                "--include-tag", "tag-to-include",
                "--exclude-tag", "tag-to-exclude"
            ])
            self.assertEqual(cmd._get_source_kwargs(), {
                "requested": spec["source"]["requested"],
                "type": spec["source"]["type"],
                "value": spec["source"]["value"],
                "tags": {
                    "include": ["tag-to-include"],
                    "exclude": ["tag-to-exclude"]
                }
            })

    def test_get_af(self):

        conf = copy.deepcopy(Configuration.DEFAULT)
        conf["specification"]["af"] = 100

        for kind, klass in self.KINDS.items():
            with mock.patch(self.CONF, conf):

                for af in (4, 6):
                    cmd = klass()
                    cmd.init_args([kind, "--af", str(af)])
                    self.assertEqual(cmd._get_af(), af)

                cmd = klass()
                cmd.init_args([kind, "--target", "1.2.3.4"])
                self.assertEqual(cmd._get_af(), 4)

                cmd = klass()
                cmd.init_args([kind, "--target", "1:2:3:4:5:6:7:8"])
                self.assertEqual(cmd._get_af(), 6)

                cmd = klass()
                cmd.init_args([kind, "--target", "1.2.3.4.5"])
                self.assertEqual(
                    cmd._get_af(),
                    conf["specification"]["af"]
                )

    def test_handle_api_error(self):

        cmd = Command()
        message = (
            "There was a problem communicating with the RIPE Atlas "
            "infrastructure.  The message given was:\n\n  {}"
        )

        with self.assertRaises(RipeAtlasToolsException) as e:
            cmd._handle_api_error("This is a plain text error")
        self.assertEqual(
            str(e.exception),
            message.format("This is a plain text error")
        )

        with self.assertRaises(RipeAtlasToolsException) as e:
            cmd._handle_api_error({"detail": "This is a formatted error"})
        self.assertEqual(
            str(e.exception),
            message.format("This is a formatted error")
        )

    def test_add_arguments(self):

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args([
                    "ping",
                    "--set-alias", "\\invalid"
                ])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                'ripe-atlas measure: error: argument --set-alias: '
                '"\\invalid" does not appear to be a valid alias.'
            )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args([
                    "ping",
                    "--renderer", "not-a-renderer"
                ])
            self.assertTrue(stderr.getvalue().split("\n")[-2].startswith(
                "ripe-atlas measure: error: argument --renderer: invalid "
                "choice: 'not-a-renderer' (choose from"
            ))

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args(["ping", "--af", "5"])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --af: invalid choice: 5 "
                "(choose from 4, 6)"
            )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args([
                    "ping",
                    "--target", "not a target"
                ])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --target: "
                "\"not a target\" does not appear to be an IP address or host "
                "name"
            )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args(["ping", "--from-area", "not an area"])
            self.assertTrue(stderr.getvalue().split("\n")[-2].startswith(
                "ripe-atlas measure: error: argument --from-area: invalid "
                "choice:"))

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                PingMeasureCommand().init_args(["ping", "--from-probes", "0,50000"])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --from-probes: 0 "
                "is lower than the minimum permitted value of 1."
            )

        for clude in ("in", "ex"):
            with capture_sys_output() as (stdout, stderr):
                for tag in ("NotATag", "tag!", "not a tag", "νοτ α ταγ"):
                    with self.assertRaises(SystemExit):
                        PingMeasureCommand().init_args([
                            "ping",
                            "--{}clude-tag".format(clude), tag
                        ])
                    self.assertEqual(
                        stderr.getvalue().split("\n")[-2],
                        'ripe-atlas measure: error: argument --{}clude-tag: '
                        '"{}" does not appear to be a valid tag.'.format(clude, tag)
                    )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                TracerouteMeasureCommand().init_args(["traceroute", "--protocol", "invalid"])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --protocol: invalid "
                "choice: 'INVALID' (choose from 'ICMP', 'UDP', 'TCP')"
            )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                DnsMeasureCommand().init_args(["dns", "--protocol", "invalid"])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --protocol: invalid "
                "choice: 'INVALID' (choose from 'UDP', 'TCP')"
            )

        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit):
                HttpMeasureCommand().init_args(
                    ["http", "--timing-verbosity", "3"])
            self.assertEqual(
                stderr.getvalue().split("\n")[-2],
                "ripe-atlas measure: error: argument --timing-verbosity: "
                "invalid choice: 3 (choose from 0, 1, 2)"
            )

        min_options = [
            ("from-measurement", ((PingMeasureCommand,), 1)),
            ("probes", ((PingMeasureCommand,), 1)),
            ("timeout", ((TracerouteMeasureCommand,), 1)),
        ]
        for option, (klasses, minimum) in min_options:
            for klass in klasses:
                with capture_sys_output() as (stdout, stderr):
                    test_value = minimum - 1
                    with self.assertRaises(SystemExit):
                        klass().init_args([
                            klass.__name__.replace("MeasureCommand", "").lower(),
                            "--{}".format(option), str(test_value)
                        ])
                    self.assertEqual(
                        stderr.getvalue().split("\n")[-2],
                        "ripe-atlas measure: error: argument --{}: "
                        "The integer must be greater than {}.".format(
                            option,
                            minimum
                        )
                    )

        min_max_options = [
            ("from-asn", (
                (PingMeasureCommand,),
                (1, 2 ** 32 - 2)
            )),
            ("paris", (
                (TracerouteMeasureCommand,),
                (0, 64)
            )),
            ("first-hop", (
                (TracerouteMeasureCommand,),
                (1, 255)
            )),
            ("max-hops", (
                (TracerouteMeasureCommand,),
                (1, 255)
            )),
            ("port", (
                (
                    TracerouteMeasureCommand,
                    SslcertMeasureCommand,
                    HttpMeasureCommand
                ),
                (1, 65535)
            )),
            ("header-bytes", ((HttpMeasureCommand,), (0, 2048))),
            ("body-bytes", ((HttpMeasureCommand,), (1, 1020048))),
            ("size", ((PingMeasureCommand,), (1, 2048))),
            ("size", ((TracerouteMeasureCommand,), (0, 2048))),
            ("packets", (
                (
                    PingMeasureCommand,
                    TracerouteMeasureCommand,
                    NtpMeasureCommand,
                ), (1, 16))),
            ("packet-interval", ((PingMeasureCommand,), (2, 30000))),
            ("timeout", ((NtpMeasureCommand,), (1, 60000))),
            ("destination-option-size", ((TracerouteMeasureCommand,), (0, 1024))),
            ("hop-by-hop-option-size", ((TracerouteMeasureCommand,), (0, 2048))),
            ("retry", ((DnsMeasureCommand,), (0, 10))),
            ("udp-payload-size", ((DnsMeasureCommand,), (512, 4096))),
        ]
        for option, (klasses, extremes) in min_max_options:
            for klass in klasses:
                lower = extremes[0] - 1
                upper = extremes[1] + 1
                for invalid in lower, upper:
                    with capture_sys_output() as (stdout, stderr):
                        try:
                            klass().init_args([
                                klass.__name__.replace("MeasureCommand", "").lower(),
                                "--{}".format(option), str(invalid)
                            ])
                        except SystemExit:
                            pass
                        else:
                            self.fail("--{} for {} should fail for value {}".format(
                                option, klass.__name__, invalid
                            ))
                        self.assertEqual(
                            stderr.getvalue().split("\n")[-2],
                            "ripe-atlas measure: error: argument --{}: The "
                            "integer must be between {} and {}.".format(
                                option, extremes[0], extremes[1]
                            )
                        )

    @mock.patch(CONF, Configuration.DEFAULT)
    def test_account_for_selected_probes(self):

        spec = Configuration.DEFAULT["specification"]

        cmd = PingMeasureCommand()
        cmd.init_args(["ping", "--target", "ripe.net"])
        cmd._account_for_selected_probes(),
        self.assertEqual(cmd.arguments.probes, spec["source"]["requested"])

        cmd = PingMeasureCommand()
        cmd.init_args(["ping", "--target", "ripe.net", "--probes", "7"])
        cmd._account_for_selected_probes(),
        self.assertEqual(cmd.arguments.probes, 7)

        cmd = PingMeasureCommand()
        cmd.init_args(["ping", "--target", "ripe.net", "--from-probes", "1,2"])
        cmd._account_for_selected_probes(),
        self.assertEqual(cmd.arguments.probes, 2)

        cmd = PingMeasureCommand()
        cmd.init_args([
            "ping",
            "--target", "ripe.net",
            "--from-probes", "1,2",
            "--probes", "7"
        ])
        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(RipeAtlasToolsException):
                cmd._account_for_selected_probes(),

    def test_set_alias(self):
        path_aliases = "ripe.atlas.tools.commands.measure.base.aliases"
        new_aliases = copy.deepcopy(AliasesDB.DEFAULT)

        with mock.patch(path_aliases, new_aliases):
            path_AliasesDB = "ripe.atlas.tools.commands.measure.base.AliasesDB"
            with mock.patch(path_AliasesDB, autospec=True) as new_AliasesDB:
                new_AliasesDB.write.return_value = True

                path_create = "ripe.atlas.tools.commands.measure.base.Command.create"
                with mock.patch(path_create) as mock_create:
                    mock_create.return_value = (
                        True,
                        {"measurements": [1234]}
                    )
                    cmd = PingMeasureCommand()
                    cmd.init_args([
                        "ping",
                        "--target",
                        "www.ripe.net",
                        "--no-report",
                        "--set-alias",
                        "PING_RIPE"
                    ])
                    cmd.run()
                    self.assertEqual(
                        new_aliases["measurement"]["PING_RIPE"],
                        1234
                    )



import os
from os.path import abspath, dirname, join
from setuptools import setup

__version__ = None
exec(open("ripe/atlas/tools/version.py").read())

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Get proper long description for package
current_dir = dirname(abspath(__file__))
description = open(join(current_dir, "README.rst")).read()
changes = open(join(current_dir, "CHANGES.rst")).read()
long_description = '\n\n'.join([description, changes])

# Get the long description from README.md
setup(
    name="ripe.atlas.tools",
    version=__version__,
    packages=["ripe", "ripe.atlas", "ripe.atlas.tools"],
    namespace_packages=["ripe", "ripe.atlas"],
    include_package_data=True,
    license="GPLv3",
    description="The official command line client for RIPE Atlas",
    long_description=long_description,
    url="https://github.com/RIPE-NCC/ripe-atlas-tools",
    download_url="https://github.com/RIPE-NCC/ripe-atlas-tools",
    author="The RIPE Atlas team",
    author_email="atlas@ripe.net",
    maintainer="The RIPE Atlas team",
    maintainer_email="atlas@ripe.net",
    install_requires=[
        "IPy",
        "python-dateutil",
        "requests>=2.7.0",
        "ripe.atlas.cousteau==1.4",
        "ripe.atlas.sagan==1.2",
        "tzlocal",
        "pyyaml",
        "pyOpenSSL>=0.13",
    ],
    tests_require=[
        "nose",
        "coverage",
        "mock",
    ],
    extras_require={
        "doc": ["sphinx", "sphinx_rtd_theme"],
        "fast": ["ujson"],
    },
    test_suite="nose.collector",
    scripts=[
        "scripts/aping",
        "scripts/atraceroute",
        "scripts/adig",
        "scripts/asslcert",
        "scripts/ahttp",
        "scripts/antp",
        "scripts/ripe-atlas",
    ],
    keywords=['RIPE', 'RIPE NCC', 'RIPE Atlas', 'Command Line'],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ...helpers.validators import ArgumentType
from ...settings import conf

from .base import Command


class TracerouteMeasureCommand(Command):

    def _upper_str(self, s):
        """
        Private method to validate specific command line arguments that
        should be provided in upper or lower case
        :param s: string
        :return: string in upper case
        """
        return s.upper()

    def add_arguments(self):

        Command.add_arguments(self)

        spec = conf["specification"]["types"]["traceroute"]

        specific = self.parser.add_argument_group(
            "Traceroute-specific Options")
        specific.add_argument(
            "--packets",
            type=ArgumentType.integer_range(minimum=1, maximum=16),
            default=spec["packets"],
            help="The number of packets sent"
        )
        specific.add_argument(
            "--size",
            type=ArgumentType.integer_range(minimum=0, maximum=2048),
            default=spec["size"],
            help="The size of packets sent"
        )
        specific.add_argument(
            "--protocol",
            type=self._upper_str,
            choices=("ICMP", "UDP", "TCP"),
            default=spec["protocol"],
            help="The protocol used."
        )
        specific.add_argument(
            "--timeout",
            type=ArgumentType.integer_range(minimum=1),
            default=spec["timeout"],
            help="The timeout per-packet"
        )
        specific.add_argument(
            "--dont-fragment",
            action="store_true",
            default=spec["dont-fragment"],
            help="Don't Fragment the packet"
        )
        specific.add_argument(
            "--paris",
            type=ArgumentType.integer_range(minimum=0, maximum=64),
            default=spec["paris"],
            help="Use Paris. Value must be between 0 and 64."
                 "If 0, a standard traceroute will be performed"
        )
        specific.add_argument(
            "--first-hop",
            type=ArgumentType.integer_range(minimum=1, maximum=255),
            default=spec["first-hop"],
            help="Value must be between 1 and 255"
        )
        specific.add_argument(
            "--max-hops",
            type=ArgumentType.integer_range(minimum=1, maximum=255),
            default=spec["max-hops"],
            help="Value must be between 1 and 255"
        )
        specific.add_argument(
            "--port",
            type=ArgumentType.integer_range(minimum=1, maximum=65535),
            default=spec["port"],
            help="Destination port, valid for TCP only"
        )
        specific.add_argument(
            "--destination-option-size",
            type=ArgumentType.integer_range(minimum=0, maximum=1024),
            default=spec["destination-option-size"],
            help="IPv6 destination option header"
        )
        specific.add_argument(
            "--hop-by-hop-option-size",
            type=ArgumentType.integer_range(minimum=0, maximum=2048),
            default=spec["hop-by-hop-option-size"],
            help=" IPv6 hop by hop option header"
        )
        specific.add_argument(
            "--duplicate-timeout",
            default=spec["duplicate-timeout"],
            type=int,
            help="Time to wait (in milliseconds) for a duplicate response "
            "after receiving the first response",
        )
        specific.add_argument(
            "--response-timeout",
            default=spec["response-timeout"],
            type=ArgumentType.integer_range(minimum=1, maximum=60000),
            help="Response timeout for one packet",
        )

    def _get_measurement_kwargs(self):

        r = Command._get_measurement_kwargs(self)

        keys = (
            "destination_option_size", "dont_fragment", "first_hop",
            "hop_by_hop_option_size", "max_hops", "packets", "paris", "port",
            "protocol", "size", "timeout"
        )
        for key in keys:
            r[key] = getattr(self.arguments, key)
        optional_keys = [
            "duplicate_timeout", "response_timeout"
        ]
        for key in optional_keys:
            val = getattr(self.arguments, key)
            if val is not None:
                r[key] = val

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import Renderer as BaseRenderer

from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from ..ipdetails import IP


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_TRACEROUTE]

    DEFAULT_SHOW_ASNS = False

    @staticmethod
    def add_arguments(parser):
        group = parser.add_argument_group(
            title="Optional arguments for traceroute renderer"
        )
        group.add_argument(
            "--traceroute-show-asns",
            help="Show Autonomous System Numbers (ASNs) in the traceroute "
                 "results.",
            action="store_true",
            default=Renderer.DEFAULT_SHOW_ASNS
        )

    def __init__(self, *args, **kwargs):
        BaseRenderer.__init__(self, *args, **kwargs)

        if "arguments" in kwargs:
            self.show_asns = kwargs["arguments"].traceroute_show_asns
        else:
            self.show_asns = Renderer.DEFAULT_SHOW_ASNS

    def on_result(self, result):

        r = ""

        for hop in result.hops:

            if hop.is_error:
                r += "{}\n".format(
                    colourise(sanitise(hop.error_message), "red"))
                continue

            name = ""
            asn = ""
            rtts = []
            for packet in hop.packets:
                name = name or packet.origin or "*"
                if self.show_asns:
                    if packet.origin and not asn:
                        asn = IP(packet.origin).asn
                if packet.rtt:
                    rtts.append("{:8} ms".format(packet.rtt))
                else:
                    rtts.append("          *")

            if not asn:
                tpl = "{hop:>3} {name:37} {rtts}\n"
            else:
                tpl = "{hop:>3} {name:28} {asn:>8} {rtts}\n"

            r += tpl.format(
                hop=hop.index,
                name=sanitise(name),
                asn="AS{}".format(asn) if asn else "",
                rtts="  ".join(rtts)
            )

        return "\n{}\n\n{}".format(
            colourise("Probe #{}".format(result.probe_id), "bold"),
            r
        )



# Copyright (c) 2017 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tzlocal import get_localzone

from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_DNS]
    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    ANSWER_COLORS = ['cyan', 'blue']

    def on_result(self, result):

        created = result.created.astimezone(get_localzone())
        probe_id = result.probe_id

        r = []
        if result.responses:
            for response in result.responses:
                r.append(self.get_formatted_response(probe_id,
                                                     created, response))
        else:
            r.append("{}{}\n".format(self.get_header(probe_id, created),
                                     colourise("No response found", "red"),
                                     ))

        return "".join(r)

    @classmethod
    def get_header(cls, probe_id, created):
        return "Probe {0:>6}: {1} ".format("#{}".format(probe_id),
                                           created.strftime(cls.TIME_FORMAT),
                                           )

    @classmethod
    def get_formatted_response(cls, probe_id, created, response):

        s = []
        answers = ""
        if response.abuf:
            header_flags = []
            for flag in ("aa", "ad", "cd", "qr", "ra", "rd",):
                if getattr(response.abuf.header, flag):
                    header_flags.append(flag)
            s.append(response.abuf.header.return_code)
            s.append(" ".join(header_flags))
            answers = cls.print_answers(response.abuf.answers)
        else:
            s.append("No abuf found")

        if response.is_error or not response.abuf:
            color = "red"
        elif len(response.abuf.answers) == 0:
            color = "yellow"
        else:
            color = "green"

        status = colourise(" ".join(s), color)
        return "". join([cls.get_header(probe_id, created),
                         status,
                         " " if answers else "",
                         answers,
                         "\n"
                         ])

    @staticmethod
    def get_rrdata(data):
        """
        Return RRdata in condensed text form.
        """
        if not data:
            return ""
        # It's too complicated to override __str__ method of all Answer
        # classes of Sagan so let's compress it text-wise instead
        r = str(data).split()
        r.pop(2)  # get rid of the class
        return sanitise(" ".join(r))

    @classmethod
    def print_answers(cls, data):
        """
        Return list of colourised condensed textual RRdata.
        """
        r = []
        for record, color in zip(data, cls.ANSWER_COLORS*len(data)):
            r.append(colourise(cls.get_rrdata(record), color))
        return "; ".join(r)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):
    """
    This is meant to be a stub example for what an aggregate renderer might look
    like. If you have ideas as to how to make this better, feel free to send
    along a pull request.
    """

    RENDERS = [BaseRenderer.TYPE_PING]

    def __init__(self):
        self.target = ""
        self.packet_loss = 0
        self.sent_packets = 0
        self.received_packets = 0
        self.rtts = []
        self.rtts_min = []
        self.rtts_max = []
        self.rtt_types_map = {
            "min": self.rtts_min,
            "max": self.rtts_max
        }

    def header(self):
        return "Collecting results...\n"

    def additional(self, results):
        self.collect_stats(results)
        self.packet_loss = self.calculate_loss()
        return self.render(
            "reports/aggregate_ping.txt",
            target=sanitise(self.target),
            sent=self.sent_packets,
            received=self.received_packets,
            packet_loss=self.packet_loss,
            min=min(self.rtts_min),
            median=self.median(),
            mean=self.mean(),
            max=max(self.rtts_max)
        )

    def collect_stats(self, results):
        """
        Calculates, stores and collects all stats we want from the given
        results.
        """
        for result in results:
            self.set_target(result)

            self.sent_packets += result.packets_sent
            self.received_packets += result.packets_received
            self.collect_min_max_rtts("min", result.rtt_min)
            self.collect_min_max_rtts("max", result.rtt_max)

            self.collect_packets_rtt(result.packets)

    def set_target(self, result):
        """Sets the target of the measurement if not set."""
        if not self.target:
            self.target = result.destination_name

    def collect_min_max_rtts(self, rtt_type, rtt):
        """
        Stores the given rtt in the corresponding list (min/max) if rtt is set.
        """
        rtt = rtt
        if not rtt:
            rtt = 0

        self.rtt_types_map[rtt_type].append(rtt)

    def collect_packets_rtt(self, packets):
        """
        Collects all the rrts of given packets and stores them
        in our rtts list.
        """
        for packet in packets:
            rtt = packet.rtt
            if not packet.rtt:
                rtt = 0
            self.rtts.append(rtt)

    def calculate_loss(self):
        """Calculates the total loss between received and sent packets."""
        if not self.sent_packets:
            return 0

        return (1 - float(self.received_packets) / self.sent_packets) * 100

    def mean(self):
        """Calculates the mean of the collected rtts"""
        return round(
            float(sum(self.rtts)) / max(len(self.rtts), 1), 3
        )

    def median(self):
        """Calculates the median of the collected rtts"""
        sorted_rtts = sorted(self.rtts)
        index = (len(self.rtts) - 1) // 2
        if len(self.rtts) % 2:
            return sorted_rtts[index]
        else:
            return (sorted_rtts[index] + sorted_rtts[index + 1]) / 2.0

    def on_result(self, result):
        return ""



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import unittest
import requests
import shutil
import tempfile

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from ripe.atlas.tools.ipdetails import IP
from ripe.atlas.tools.cache import LocalCache


class FakeCache(LocalCache):


    def __init__(self):
        self.paths = []
        LocalCache.__init__(self)

    def _get_or_create_db_path(self):
        path = os.path.join(tempfile.mkdtemp(), "ripe.atlas.tool.unittest")
        self.paths.append(path)
        return path

    def test_cleanup(self):
        for path in self.paths:
            shutil.rmtree(os.path.dirname(path))


fake_cache = FakeCache()


class FakeResponse(object):
    def __init__(self, json_return={}, ok=True):
        self.json_return = json_return
        self.ok = ok
        self.text = "testing"

    def json(self):
        return self.json_return


class FakeErrorResponse(FakeResponse):
    def json(self):
        raise ValueError("json breaks")


class TestIPDetails(unittest.TestCase):

    IP = '193.0.6.1'
    ASN = '3333'
    HOLDER = 'RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL'
    SAME_PREFIX_IP = '193.0.6.2'
    PREFIX = '193.0.0.0/21'
    SAME_AS_DIFFERENT_PREFIX_IP = '193.0.22.1'
    NOT_ANNOUNCED_IP = '80.81.192.1'
    MOCK_RESULTS = {
        IP: {"status": "ok", "server_id": "stat-app2", "cached": False, "status_code": 200, "time": "2015-10-12T15:30:00.113317", "messages": [["warning", "Given resource is not announced but result has been aligned to first-level less-specific (193.0.0.0/21)."]], "version": "1.3", "data_call_status": "supported - connecting to ursa", "see_also": [], "process_time": 561, "query_id": "196d2754-70f6-11e5-b8ba-782bcb346712", "data": {"query_time": "2015-10-12T08:00:00", "is_less_specific": True, "resource": "193.0.0.0/21", "actual_num_related": 0, "num_filtered_out": 0, "asns": [{"holder": "RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL", "asn": 3333}], "announced": True, "related_prefixes": [], "type": "prefix", "block": {"resource": "193.0.0.0/8", "name": "IANA IPv4 Address Space Registry", "desc": "RIPE NCC (Status: ALLOCATED)"}}},
        SAME_AS_DIFFERENT_PREFIX_IP: {"status": "ok", "server_id": "stat-app2", "cached": False, "status_code": 200, "time": "2015-10-12T15:32:25.778643", "messages": [["warning", "Given resource is not announced but result has been aligned to first-level less-specific (193.0.22.0/23)."]], "version": "1.3", "data_call_status": "supported - connecting to ursa", "see_also": [], "process_time": 818, "query_id": "7018b6cc-70f6-11e5-8bf8-782bcb346712", "data": {"query_time": "2015-10-12T08:00:00", "is_less_specific": True, "resource": "193.0.22.0/23", "actual_num_related": 0, "num_filtered_out": 0, "asns": [{"holder": "RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL", "asn": 3333}], "announced": True, "related_prefixes": [], "type": "prefix", "block": {"resource": "193.0.0.0/8", "name": "IANA IPv4 Address Space Registry", "desc": "RIPE NCC (Status: ALLOCATED)"}}},
        NOT_ANNOUNCED_IP: {"status": "ok", "server_id": "stat-app2", "cached": False, "status_code": 200, "time": "2015-10-12T15:33:58.911309", "messages": [["info", "2 routes were filtered due to low visibility (min peers:3)."]], "version": "1.3", "data_call_status": "supported - connecting to ursa", "see_also": [], "process_time": 462, "query_id": "a7d1daee-70f6-11e5-aaec-782bcb346712", "data": {"query_time": "2015-10-12T08:00:00", "is_less_specific": False, "resource": "80.81.192.1", "actual_num_related": 0, "num_filtered_out": 2, "asns": [], "announced": False, "related_prefixes": [], "type": "prefix", "block": {"resource": "80.0.0.0/8", "name": "IANA IPv4 Address Space Registry", "desc": "RIPE NCC (Status: ALLOCATED)"}}}
    }

    def setUp(self):
        fake_cache.clear()

        self.mock_cache = mock.patch(
            "ripe.atlas.tools.ipdetails.cache", wraps=fake_cache
        ).start()
        self.mock_get = mock.patch(
            'ripe.atlas.tools.ipdetails.requests.get'
        ).start()
        self.mock_get.return_value = FakeResponse(
            json_return=self.MOCK_RESULTS[self.IP]
        )

    def tearDown(self):
        mock.patch.stopall()

    @classmethod
    def tearDownClass(cls):
        fake_cache._db.close()
        fake_cache.test_cleanup()

    def test_loopback4(self):
        """IPv4 loopback address"""
        det = IP("127.0.0.1")
        self.assertEquals(det.asn, None)
        self.assertEquals(det.is_querable(), False)
        # no query to stat
        self.assertEquals(self.mock_get.call_count, 0)
        # no access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 0)

    def test_loopback6(self):
        """IPv6 loopback address"""
        det = IP("::1")
        self.assertEquals(det.asn, None)
        self.assertEquals(det.is_querable(), False)
        # no query to stat
        self.assertEquals(self.mock_get.call_count, 0)
        # no access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 0)

    def test_nocache(self):
        """No cache"""
        det = IP(self.IP)

        self.assertEquals(det.asn, self.ASN)
        self.assertEquals(det.prefix, self.PREFIX)
        self.assertEquals(det.holder, self.HOLDER)
        # query to stat
        self.assertEquals(self.mock_get.call_count, 1)
        # access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 1)
        # access to cache set
        self.assertEquals(self.mock_cache.set.call_count, 2)

    def test_fakecache_sameip(self):
        """Fake cache, same IP"""
        det = IP(self.IP)
        det = IP(self.IP)

        self.assertEquals(det.asn, self.ASN)
        # query to stat
        self.assertEquals(self.mock_get.call_count, 1)
        # access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 2)
        # access to cache set
        self.assertEquals(self.mock_cache.set.call_count, 2)

    def test_fakecache_sameprefix(self):
        """Fake cache, same prefix"""
        det1 = IP(self.IP)
        det2 = IP(self.SAME_PREFIX_IP)

        self.assertEquals(det1.asn, self.ASN)
        self.assertEquals(det2.asn, det1.asn)
        # query to stat
        self.assertEquals(self.mock_get.call_count, 1)
        # access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 3)
        # access to cache set
        self.assertEquals(self.mock_cache.set.call_count, 3)

    def test_fakecache_diffprefix(self):
        """Fake cache, same AS, different prefix"""
        self.mock_get.return_value = FakeResponse(
            json_return=self.MOCK_RESULTS[self.IP]
        )
        det1 = IP(self.IP)
        self.mock_get.return_value = FakeResponse(
            json_return=self.MOCK_RESULTS[self.SAME_AS_DIFFERENT_PREFIX_IP]
        )
        det2 = IP(self.SAME_AS_DIFFERENT_PREFIX_IP)

        self.assertEquals(det1.asn, self.ASN)
        self.assertEquals(det2.asn, det1.asn)
        self.assertEquals(self.mock_get.call_count, 2)
        # access to cache get
        self.assertEquals(self.mock_cache.get.call_count, 3)
        # access to cache set
        self.assertEquals(self.mock_cache.set.call_count, 4)

    def test_fakecache_notannounced(self):
        """Fake cache, IP not announced"""
        self.mock_get.return_value = FakeResponse(
            json_return=self.MOCK_RESULTS[self.NOT_ANNOUNCED_IP]
        )
        det = IP(self.NOT_ANNOUNCED_IP)

        self.assertEquals(det.asn, None)
        self.assertEquals(self.mock_get.call_count, 1)

        IP(self.NOT_ANNOUNCED_IP)

        # now it should be cached
        self.assertEquals(det.asn, None)
        self.assertEquals(self.mock_get.call_count, 2)

    def test_valid_query_stat(self):
        """Test case for valid stat response"""
        ip = IP(self.IP)
        self.assertEqual(
            ip.query_stat(),
            {
                'Prefix': '193.0.0.0/21',
                'Holder': 'RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL',
                'ASN': '3333'
            }
        )

    def test_invalid_query_stat(self):
        """Test case where stat returns not ok status"""
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "notok"}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        self.mock_get.return_value = FakeResponse(
            json_return={}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

    def test_invalid_query_stat1(self):
        """Test case where stat returns not valid data structure"""
        # no data at all
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "ok"}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        # data is dict but no asns key
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "ok", "data": {"bla": "bla"}}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        # data is not dict
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "ok", "data": [1, 2]}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        # asns is not list
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "ok", "data": {"asns": {}}}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        # asns is empty
        self.mock_get.return_value = FakeResponse(
            json_return={"status": "ok", "data": {"asns": [{}]}}
        )
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

    def test_invalid_query_stat2(self):
        """Test case where stat returns exception"""
        self.mock_get.side_effect = requests.exceptions.RequestException
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

        self.mock_get.return_value = FakeErrorResponse()
        ip = IP(self.IP)
        self.assertEqual(ip.query_stat(), {})

    def test_update_cache(self):
        """Test case where we store both prefix/address"""
        details = {
            'Prefix': '193.0.0.0/21',
            'Holder': 'RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL',
            'ASN': '3333'
        }
        IP(self.IP)
        self.assertEquals(self.mock_cache.set.call_count, 2)
        self.assertEquals(self.mock_cache.get("IPDetails:193.0.6.1"), details)
        self.assertEquals(self.mock_cache.get(
            "IPDetailsPrefix:193.0.0.0/21"), details
        )

    def test_update_cache1(self):
        """Test case where we store only address"""
        details = {
            'Prefix': '193.0.0.0/21',
            'Holder': 'RIPE-NCC-AS Reseaux IP Europeens Network Coordination Centre (RIPE NCC),NL',
            'ASN': '3333'
        }
        self.mock_cache.set("IPDetailsPrefix:193.0.0.0/21", details, 1)
        IP(self.IP)
        # we already called it above, so it should be 2 by bow
        self.assertEquals(self.mock_cache.set.call_count, 2)
        self.assertEquals(self.mock_cache.get("IPDetails:193.0.6.1"), details)
        self.assertEquals(self.mock_cache.get("IPDetailsPrefix:193.0.0.0/21"), details)

    def test_get_from_cache_prefix(self):
        """Test case where we have a matching prefix in cache"""
        details = {
            'Prefix': '193.0.0.0/21',
            'Holder': 'test',
            'ASN': 'test'
        }
        self.mock_cache.set("IPDetailsPrefix:193.0.0.0/20", details, 1)
        ip = IP(self.IP)
        self.assertTrue(ip.cached_prefix_found)
        self.assertEquals(ip.asn, "test")
        self.assertEquals(ip.holder, "test")
        self.assertEquals(ip.get_from_cached_prefix(), details)

    def test_get_from_cache_prefix1(self):
        """Test case where we dont' have a matching prefix in cache"""
        ip = IP(self.IP)
        # clear out db to test specific function
        self.mock_cache.clear()
        self.assertFalse(ip.cached_prefix_found)
        self.assertEquals(ip.get_from_cached_prefix(), None)

    def test_is_querable(self):
        """Test case where IP is quearable"""
        ip = IP(self.IP)
        self.assertTrue(ip.is_querable())

    def test_is_querable1(self):
        """Test case where IP is not quearable"""
        ip = IP("127.0.0.1")
        self.assertFalse(ip.is_querable())



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import os

from ..exceptions import RipeAtlasToolsException
from ..settings import Configuration, conf
from .base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "configure"

    EDITOR = os.environ.get("EDITOR", "/usr/bin/vim")
    DESCRIPTION = "Adjust or initialize configuration options"
    EXTRA_DESCRIPTION = (
        "As an alternative to this command, you can just create/edit {}"
        .format(Configuration.USER_RC)
    )

    def add_arguments(self):
        self.parser.add_argument(
            "--editor",
            action="store_true",
            help="Invoke {0} to edit the configuration directly".format(
                self.EDITOR)
        )
        self.parser.add_argument(
            "--set",
            action="store",
            help="Permanently set a configuration value so it can be used in "
                 "the future.  Example: --set authorisation.create=MY_API_KEY"
        )
        self.parser.add_argument(
            "--init",
            action="store_true",
            help="Create a configuration file and save it into your home "
                 "directory at: {}".format(Configuration.USER_RC)
        )

    def run(self):

        if not self.arguments.init:
            if not self.arguments.editor:
                if not self.arguments.set:
                    raise RipeAtlasToolsException(
                        "Run this with --help for more information")

        self._create_if_necessary()

        if self.arguments.editor:
            os.system("{0} {1}".format(self.EDITOR, Configuration.USER_RC))

        if self.arguments.init or self.arguments.editor:
            return self.ok(
                "Configuration file writen to {}".format(Configuration.USER_RC))

        if self.arguments.set:
            if "=" not in self.arguments.set:
                raise RipeAtlasToolsException(
                    "Invalid format. Execute with --help for more information.")
            path, value = self.arguments.set.split("=")
            self.set(path.split("."), value)

    def set(self, path, value):
        if path[:2] == ['authorisation', 'fetch_aliases']:
            if len(path) > 3:
                raise RipeAtlasToolsException(
                    'Invalid alias for a fetch API key: it must be in the '
                    'format authorisation.fetch.some-alias=MY_API_KEY')

            if 'fetch_aliases' not in conf['authorisation']:
                conf['authorisation']['fetch_aliases'] = {}
            if conf['authorisation']['fetch_aliases'] is None:
                conf['authorisation']['fetch_aliases'] = {}

            alias = path[2]

            if alias not in conf['authorisation']['fetch_aliases']:
                conf['authorisation']['fetch_aliases'][alias] = None

            required_type = str
        else:
            try:
                required_type = type(self._get_from_dict(conf, path))
            except KeyError:
                raise RipeAtlasToolsException(
                    'Invalid configuration key: "{}"'.format(".".join(path)))

        if value.isdigit():
            value = int(value)

        if not isinstance(value, required_type):
            raise RipeAtlasToolsException(
                'Invalid configuration value: "{}". You must supply a {} for '
                'this key'.format(value, required_type.__name__)
            )

        self._set_in_dict(conf, path, value)

        Configuration.write(conf)

    @staticmethod
    def _create_if_necessary():

        if os.path.exists(Configuration.USER_RC):
            return

        if not os.path.exists(Configuration.USER_CONFIG_DIR):
            os.makedirs(Configuration.USER_CONFIG_DIR)

        Configuration.write(conf)

    @staticmethod
    def _get_from_dict(data, path):
        return functools.reduce(lambda d, k: d[k], path, data)

    @classmethod
    def _set_in_dict(cls, data, path, value):
        cls._get_from_dict(data, path[:-1])[path[-1]] = value

    @staticmethod
    def cast_value(value):

        # Booleans are a pain in the ass to cast
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False

        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return str(value)



import os.path
import unittest
import tempfile
import shutil
import sys

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

# Python 2.7 does have io.StringIO but StringIO. is more liberal regarding str
# versus unicode inputs to write()
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from ripe.atlas.tools.commands.base import Command


USER_COMMAND = """
from ripe.atlas.tools.commands.base import Command as BaseCommand

class Command(BaseCommand):
    NAME = 'user-command-1'
"""


class TestCommandLoading(unittest.TestCase):
    expected_builtins = [
        "configure",
        "alias",
        "go",
        "measure",
        "measurement-info",
        "measurement-search",
        "probe-info",
        "probe-search",
        "report",
        "shibboleet",
        "stream",
    ]

    def setUp(self):
        # Create a directory for storing user commands and insert the dummy
        # command
        self.user_command_path = tempfile.mkdtemp()
        with open(
            os.path.join(self.user_command_path, "user_command_1.py"),
            "w"
        ) as f:
            f.write(USER_COMMAND)

    def tearDown(self):
        shutil.rmtree(self.user_command_path)

    @mock.patch(
        "ripe.atlas.tools.commands.base.Command._get_user_command_path",
        return_value=None,
    )
    def test_command_loading(self, _get_user_command_path):
        _get_user_command_path.return_value = self.user_command_path

        # For some reason the command classes are initialized in some envs
        # but not in others at this point
        if Command._commands:
            Command._commands.clear()
        Command._load_commands()

        available_commands = Command.get_available_commands()

        # Check that we have the command list that we expect
        self.assertEquals(
            sorted(available_commands),
            sorted(
                [b.replace('-', '_') for b in self.expected_builtins] +
                ["user_command_1"]
            ),
        )

        # Check that we can load (i.e. import) every builtin command
        for expected_builtin in self.expected_builtins:
            self.assertIn(expected_builtin.replace("-", "_"), available_commands)
            cmd_cls = Command.load_command_class(expected_builtin)
            self.assertIsNotNone(cmd_cls)
            self.assertEquals(cmd_cls.get_name(), expected_builtin)

        # Check that we can load the user command
        user_cmd_cls = Command.load_command_class("user-command-1")
        self.assertIsNotNone(user_cmd_cls)
        self.assertEquals(user_cmd_cls.get_name(), "user-command-1")

        # Check that load_command_class() returns None for junk commands
        unexpected_cmd = Command.load_command_class("no-such-command")
        self.assertIsNone(unexpected_cmd)

    def test_deprecated_aliases(self):
        aliases = [
            ("measurement", "measurement-info"),
            ("measurements", "measurement-search"),
            ("probe", "probe-info"),
            ("probes", "probe-search"),
        ]

        # Check that each alias is loaded correctly and outputs a warning
        stderr = sys.stderr
        sys.stderr = StringIO()
        try:
            for alias, cmd_name in aliases:
                sys.stderr.truncate()
                cmd_cls = Command.load_command_class(alias)
                self.assertIn(
                    "{} is a deprecated alias for {}".format(alias, cmd_name),
                    sys.stderr.getvalue(),
                )
                self.assertIsNotNone(cmd_cls)
                self.assertEquals(cmd_cls.get_name(), cmd_name)
        finally:
            sys.stderr = stderr
